@@ -350,19 +350,24 @@ static struct arg_info _arg_info[] = {
 	  "Shared key file (default=" DEFAULT_KEY_FILE ")",
 	  assign_key },
 
-	{ 'o', "-o <operation>", "option",
+	{ '\xff', NULL, "option",
 	  0, "string", "reboot",
 	  "Fencing operation (null, off, [reboot])",
 	  assign_op },
 
+	{ 'o', "-o <action>", "action",
+	  0, "string", "reboot",
+	  "Fencing action (null, off, [reboot])",
+	  assign_op },
+
 	{ 'H', "-H <domain>", "domain",
 	  1, "string", NULL,
-	  "Xen host (domain name) to fence",
+	  "Virtual machine (domain name) to fence",
 	  assign_domain },
 
 	{ 'u', "-u", "use_uuid",
 	  0, "string", NULL,
-	  "Treat <domain> as UUID instead of domain name",
+	  "Treat 'domain' as UUID instead of domain name",
 	  assign_uuid_lookup },
 
 	{ 't', "-t <timeout>", "timeout",
