@@ -270,6 +270,37 @@ do_read_readi(struct file *file, char *buf, size_t size, loff_t *offset,
 	return count;
 }
 
+/**
+ * grope_mapping - feel up a mapping that needs to be written
+ * @buf: the start of the memory to be written
+ * @size: the size of the memory to be written
+ *
+ * We do this after acquiring the locks on the mapping,
+ * but before starting the write transaction.  We need to make
+ * sure that we don't cause recursive transactions if blocks
+ * need to be allocated to the file backing the mapping.
+ *
+ * Returns: errno
+ */
+
+static int
+grope_mapping(char *buf, size_t size)
+{
+	unsigned long start = (unsigned long)buf;
+	unsigned long stop = start + size;
+	char c;
+
+	while (start < stop) {
+		if (copy_from_user(&c, (char *)start, 1))
+			return -EFAULT;
+
+		start += PAGE_CACHE_SIZE;
+		start &= PAGE_CACHE_MASK;
+	}
+
+	return 0;
+}
+
 /**
  * do_read_direct - Read bytes from a file
  * @file: The file to read from
@@ -306,6 +337,12 @@ do_read_direct(struct file *file, char *buf, size_t size, loff_t *offset,
 
 	gfs_holder_init(ip->i_gl, state, flags, &ghs[num_gh]);
 
+	if (num_gh && atomic_read(&current->mm->mm_users) > 1) {
+		error = grope_mapping(buf, size);
+		if (error)
+			goto out;
+	}
+
 	error = gfs_glock_nq_m(num_gh + 1, ghs);
 	if (error)
 		goto out;
@@ -315,7 +352,7 @@ do_read_direct(struct file *file, char *buf, size_t size, loff_t *offset,
 		goto out_gunlock;
 
 	if (gfs_is_stuffed(ip)) {
-		size_t mask = bdev_hardsect_size(inode->i_sb->s_bdev) - 1;
+		size_t mask = bdev_logical_block_size(inode->i_sb->s_bdev) - 1;
 
 		if (((*offset) & mask) || (((unsigned long)buf) & mask))
 			goto out_gunlock;
@@ -369,6 +406,12 @@ do_read_buf(struct file *file, char *buf, size_t size, loff_t *offset,
 
 	gfs_holder_init(ip->i_gl, LM_ST_SHARED, GL_ATIME, &ghs[num_gh]);
 
+	if (num_gh && atomic_read(&current->mm->mm_users) > 1) {
+		error = grope_mapping(buf, size);
+		if (error)
+			goto out;
+	}
+
 	error = gfs_glock_nq_m_atime(num_gh + 1, ghs);
 	if (error)
 		goto out;
@@ -438,37 +481,6 @@ gfs_aio_read(struct kiocb *iocb, const struct iovec *iov, unsigned long count,
 	return(__gfs_read(filp, iov->iov_base, iov->iov_len, &iocb->ki_pos, iocb));
 }
 
-/**
- * grope_mapping - feel up a mapping that needs to be written
- * @buf: the start of the memory to be written
- * @size: the size of the memory to be written
- *
- * We do this after acquiring the locks on the mapping,
- * but before starting the write transaction.  We need to make
- * sure that we don't cause recursive transactions if blocks
- * need to be allocated to the file backing the mapping.
- *
- * Returns: errno
- */
-
-static int
-grope_mapping(char *buf, size_t size)
-{
-	unsigned long start = (unsigned long)buf;
-	unsigned long stop = start + size;
-	char c;
-
-	while (start < stop) {
-		if (copy_from_user(&c, (char *)start, 1))
-			return -EFAULT;
-
-		start += PAGE_CACHE_SIZE;
-		start &= PAGE_CACHE_MASK;
-	}
-
-	return 0;
-}
-
 /**
  * gfs_file_aio_write_nolock - Call vfs aio layer to write bytes to a file
  * @file: The file to write to
@@ -671,6 +683,12 @@ do_write_direct(struct file *file, char *buf, size_t size, loff_t *offset,
  restart:
 	gfs_holder_init(ip->i_gl, state, 0, &ghs[num_gh]);
 
+	if (num_gh && atomic_read(&current->mm->mm_users) > 1) {
+		error = grope_mapping(buf, size);
+		if (error)
+			goto out;
+	}
+
 	error = gfs_glock_nq_m(num_gh + 1, ghs);
 	if (error)
 		goto out;
@@ -959,6 +977,12 @@ do_write_buf(struct file *file,
 
 	gfs_holder_init(ip->i_gl, LM_ST_EXCLUSIVE, 0, &ghs[num_gh]);
 
+	if (num_gh && atomic_read(&current->mm->mm_users) > 1) {
+		error = grope_mapping(buf, size);
+		if (error)
+			goto out;
+	}
+
 	error = gfs_glock_nq_m(num_gh + 1, ghs);
 	if (error)
 		goto out;
@@ -1132,6 +1156,29 @@ filldir_reg_func(void *opaque,
 	return 0;
 }
 
+static inline int
+high_stat_rate(struct gfs_inode *ip)
+{
+	struct timeval now, diff;
+	unsigned long milli = 0;
+	unsigned int rate = 0; /* stats per second in this dir */
+
+	do_gettimeofday(&now);
+	diff.tv_sec = now.tv_sec - ip->i_dir_stat_st.tv_sec;
+	diff.tv_usec = now.tv_usec - ip->i_dir_stat_st.tv_usec;
+	if (diff.tv_usec < 0) {
+		diff.tv_usec += 1000000;
+		diff.tv_sec--;
+	}
+	milli = (diff.tv_sec * 1000) + (diff.tv_usec / 1000);
+	if (milli)
+		rate = (ip->i_dir_stats * 1000) / milli;
+
+	if (rate > 10) /* More than 10 stats/sec */
+		return 1;
+	return 0;
+}
+
 /**
  * readdir_reg - Read directory entries from a directory
  * @file: The directory to read from
@@ -1151,10 +1198,14 @@ readdir_reg(struct file *file, void *dirent, filldir_t filldir)
 	int error;
 
 	fdr.fdr_sbd = dip->i_sbd;
-	fdr.fdr_prefetch = TRUE;
+	fdr.fdr_prefetch = high_stat_rate(dip) ? TRUE : FALSE;
 	fdr.fdr_filldir = filldir;
 	fdr.fdr_opaque = dirent;
 
+	/* reset stat counter and timestamp */
+	dip->i_dir_stats = 0;
+	do_gettimeofday(&dip->i_dir_stat_st);
+
 	gfs_holder_init(dip->i_gl, LM_ST_SHARED, GL_ATIME, &d_gh);
 	error = gfs_glock_nq_atime(&d_gh);
 	if (error) {
@@ -1571,6 +1622,8 @@ gfs_fsync(struct file *file, struct dentry *dentry, int datasync)
 			};
 			error = sync_inode(inode, &wbc);
 		}
+		if (gfs_is_stuffed(ip))
+			gfs_log_flush_glock(ip->i_gl);
 	}
 
 	gfs_glock_dq_uninit(&i_gh);
@@ -1603,6 +1656,12 @@ gfs_lock(struct file *file, int cmd, struct file_lock *fl)
 	if ((ip->i_di.di_mode & (S_ISGID | S_IXGRP)) == S_ISGID)
 		return -ENOLCK;
 
+	if (cmd == F_CANCELLK) {
+		/* Hack: */
+		cmd = F_SETLK;
+		fl->fl_type = F_UNLCK;
+	}
+
 	if (IS_GETLK(cmd))
 		return gfs_lm_plock_get(sdp, &name, file, fl);
 	else if (fl->fl_type == F_UNLCK)
@@ -1778,6 +1837,24 @@ struct file_operations gfs_file_fops = {
 	.flock = gfs_flock,
 };
 
+struct file_operations gfs_file_fops_jdata = {
+	.llseek = gfs_llseek,
+	.read = gfs_read,
+	.write = gfs_write,
+	.aio_read = gfs_aio_read,
+	.aio_write = gfs_aio_write,
+	.ioctl = gfs_ioctl,
+#ifdef CONFIG_COMPAT
+	.compat_ioctl   = gfs_compat_ioctl,
+#endif
+	.mmap = gfs_mmap,
+	.open = gfs_open,
+	.release = gfs_close,
+	.fsync = gfs_fsync,
+	.lock = gfs_lock,
+	.flock = gfs_flock,
+};
+
 struct file_operations gfs_dir_fops = {
 	.readdir = gfs_readdir,
 	.ioctl = gfs_ioctl,
@@ -1808,6 +1885,22 @@ struct file_operations gfs_file_fops_nolock = {
 	.splice_read = gfs_splice_read,
 };
 
+struct file_operations gfs_file_fops_nolock_jdata = {
+	.llseek = gfs_llseek,
+	.read = gfs_read,
+	.write = gfs_write,
+	.aio_read = gfs_aio_read,
+	.aio_write = gfs_aio_write,
+	.ioctl = gfs_ioctl,
+#ifdef CONFIG_COMPAT
+	.compat_ioctl = gfs_compat_ioctl,
+#endif
+	.mmap = gfs_mmap,
+	.open = gfs_open,
+	.release = gfs_close,
+	.fsync = gfs_fsync,
+};
+
 struct file_operations gfs_dir_fops_nolock = {
 	.readdir = gfs_readdir,
 	.ioctl = gfs_ioctl,
