@@ -12,6 +12,7 @@
 #include <linux/xattr.h>
 #include <linux/posix_acl.h>
 #include <linux/security.h>
+#include <linux/cred.h>
 
 #include "gfs.h"
 #include "acl.h"
@@ -139,7 +140,7 @@ gfs_create(struct inode *dir, struct dentry *dentry,
 
 	if (!inode)
 		error = -ENOMEM;
-	else
+	else if (new)
 		error = gfs_security_init(dip, ip);
 
 	gfs_glock_dq_uninit(&d_gh);
@@ -191,11 +192,11 @@ lookup_cdpn_sub_at(struct gfs_sbd *sdp, struct dentry *dentry)
 	else if (gfs_filecmp(&dentry->d_name, "@uid", 4))
 		new = lookup_one_len(buf,
 				     parent,
-				     sprintf(buf, "%u", current->fsuid));
+				     sprintf(buf, "%u", current_fsuid()));
 	else if (gfs_filecmp(&dentry->d_name, "@gid", 4))
 		new = lookup_one_len(buf,
 				     parent,
-				     sprintf(buf, "%u", current->fsgid));
+				     sprintf(buf, "%u", current_fsgid()));
 	else if (gfs_filecmp(&dentry->d_name, "@sys", 4))
 		new = lookup_one_len(buf,
 				     parent,
@@ -250,11 +251,11 @@ lookup_cdpn_sub_brace(struct gfs_sbd *sdp, struct dentry *dentry)
 	else if (gfs_filecmp(&dentry->d_name, "{uid}", 5))
 		new = lookup_one_len(buf,
 				     parent,
-				     sprintf(buf, "%u", current->fsuid));
+				     sprintf(buf, "%u", current_fsuid()));
 	else if (gfs_filecmp(&dentry->d_name, "{gid}", 5))
 		new = lookup_one_len(buf,
 				     parent,
-				     sprintf(buf, "%u", current->fsgid));
+				     sprintf(buf, "%u", current_fsgid()));
 	else if (gfs_filecmp(&dentry->d_name, "{sys}", 5))
 		new = lookup_one_len(buf,
 				     parent,
@@ -1493,8 +1494,8 @@ gfs_setattr(struct dentry *dentry, struct iattr *attr)
 static int
 gfs_getattr(struct vfsmount *mnt, struct dentry *dentry, struct kstat *stat)
 {
-	struct inode *inode = dentry->d_inode;
-	struct gfs_inode *ip = get_v2ip(inode);
+	struct inode *inode = dentry->d_inode, *p_inode = NULL;
+	struct gfs_inode *ip = get_v2ip(inode), *pi = NULL;
 	struct gfs_holder gh;
 	int error;
 
@@ -1503,6 +1504,15 @@ gfs_getattr(struct vfsmount *mnt, struct dentry *dentry, struct kstat *stat)
 	error = gfs_glock_nq_init(ip->i_gl, LM_ST_SHARED, LM_FLAG_ANY, &gh);
 	if (!error) {
 		generic_fillattr(inode, stat);
+		if (S_ISREG(inode->i_mode) && dentry->d_parent 
+		    && dentry->d_parent->d_inode) {
+			p_inode = igrab(dentry->d_parent->d_inode);
+			if (p_inode) {
+				pi = get_v2ip(p_inode);
+				pi->i_dir_stats++;
+				iput(p_inode);
+			}
+		}
 		gfs_glock_dq_uninit(&gh);
 	}
 
