@@ -1,25 +1,27 @@
 #include <stdint.h>
 #include <string.h>
-#include <linux_endian.h>
+#include <libintl.h>
+#define _(String) gettext(String)
 
 #include "libgfs2.h"
 #include "fsck.h"
+#include "eattr.h"
 
 static int clear_blk_nodup(struct gfs2_sbd *sbp, uint64_t block)
 {
 	struct gfs2_block_query q;
 
-	if(gfs2_block_check(bl, block, &q)) {
+	if(gfs2_block_check(sbp, bl, block, &q)) {
 		stack;
 		return -1;
 	}
 
 	if(q.dup_block) {
-		log_debug("Not clearing block with marked as a duplicate\n");
+		log_debug( _("Not clearing block with marked as a duplicate\n"));
 		return 1;
 	}
 
-	gfs2_block_set(bl, block, gfs2_block_free);
+	gfs2_block_set(sbp, bl, block, gfs2_block_free);
 
 	return 0;
 
@@ -27,22 +29,20 @@ static int clear_blk_nodup(struct gfs2_sbd *sbp, uint64_t block)
 
 int clear_eattr_indir(struct gfs2_inode *ip, uint64_t block,
 		      uint64_t parent, struct gfs2_buffer_head **bh,
-		      void *private)
+		      enum update_flags *want_updated, void *private)
 {
+	*want_updated = not_updated;
 	return clear_blk_nodup(ip->i_sbd, block);
 }
 
-
 int clear_eattr_leaf(struct gfs2_inode *ip, uint64_t block,
 		     uint64_t parent, struct gfs2_buffer_head **bh,
-		     void *private)
+		     enum update_flags *want_updated, void *private)
 {
-
+	*want_updated = not_updated;
 	return clear_blk_nodup(ip->i_sbd, block);
-
 }
 
-
 int clear_eattr_entry (struct gfs2_inode *ip,
 		       struct gfs2_buffer_head *leaf_bh,
 		       struct gfs2_ea_header *ea_hdr,
@@ -77,8 +77,8 @@ int clear_eattr_entry (struct gfs2_inode *ip,
 		if(max_ptrs > ea_hdr->ea_num_ptrs) {
 			return 1;
 		} else {
-			log_debug("  Pointers Required: %d\n"
-				  "  Pointers Reported: %d\n",
+			log_debug( _("  Pointers Required: %d\n"
+				  "  Pointers Reported: %d\n"),
 				  max_ptrs,
 				  ea_hdr->ea_num_ptrs);
 		}
@@ -89,11 +89,14 @@ int clear_eattr_entry (struct gfs2_inode *ip,
 }
 
 int clear_eattr_extentry(struct gfs2_inode *ip, uint64_t *ea_data_ptr,
-			 struct gfs2_buffer_head *leaf_bh, struct gfs2_ea_header *ea_hdr,
-			 struct gfs2_ea_header *ea_hdr_prev, void *private)
+			 struct gfs2_buffer_head *leaf_bh,
+			 struct gfs2_ea_header *ea_hdr,
+			 struct gfs2_ea_header *ea_hdr_prev,
+			 enum update_flags *want_updated, void *private)
 {
 	uint64_t block = be64_to_cpu(*ea_data_ptr);
 
+	*want_updated = not_updated;
 	return clear_blk_nodup(ip->i_sbd, block);
 
 }
