@@ -12,8 +12,18 @@
 #include <linux/types.h>
 #include "libgfs2.h"
 
+static __inline__ uint64_t *metapointer(struct gfs2_buffer_head *bh,
+					unsigned int height,
+					struct metapath *mp)
+{
+	unsigned int head_size = (height > 0) ?
+		sizeof(struct gfs2_meta_header) : sizeof(struct gfs2_dinode);
+
+	return ((uint64_t *)(bh->b_data + head_size)) + mp->mp_list[height];
+}
+
 /* Detect directory is a stuffed inode */
-int inode_is_stuffed(struct gfs2_inode *ip)
+static int inode_is_stuffed(struct gfs2_inode *ip)
 {
 	return !ip->i_di.di_height;
 }
@@ -22,22 +32,26 @@ struct gfs2_inode *inode_get(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh)
 {
 	struct gfs2_inode *ip;
 
-	zalloc(ip, sizeof(struct gfs2_inode));
+	ip = calloc(1, sizeof(struct gfs2_inode));
+	if (ip == NULL) {
+		fprintf(stderr, "Out of memory in %s\n", __FUNCTION__);
+		exit(-1);
+	}
 	gfs2_dinode_in(&ip->i_di, bh->b_data);
 	ip->i_bh = bh;
 	ip->i_sbd = sdp;
 	return ip;
 }
 
-void inode_put(struct gfs2_inode *ip, enum update_flags updated)
+void inode_put(struct gfs2_inode *ip, enum update_flags is_updated)
 {
-	if (updated)
+	if (is_updated)
 		gfs2_dinode_out(&ip->i_di, ip->i_bh->b_data);
-	brelse(ip->i_bh, updated);
+	brelse(ip->i_bh, is_updated);
 	free(ip);
 }
 
-uint64_t blk_alloc_i(struct gfs2_sbd *sdp, unsigned int type)
+static uint64_t blk_alloc_i(struct gfs2_sbd *sdp, unsigned int type)
 {
 	osi_list_t *tmp, *head;
 	struct rgrp_list *rl = NULL;
@@ -79,7 +93,7 @@ uint64_t blk_alloc_i(struct gfs2_sbd *sdp, unsigned int type)
 	die("allocation is broken (1): %"PRIu64" %u\n",
 	    (uint64_t)rl->ri.ri_addr, rl->rg.rg_free);
 
- found:
+found:
 	if (bn >= ri->ri_bitbytes * GFS2_NBBY)
 		die("allocation is broken (2): %u %u %"PRIu64" %u\n",
 		    bn, ri->ri_bitbytes * GFS2_NBBY,
@@ -152,7 +166,7 @@ buffer_copy_tail(struct gfs2_sbd *sdp,
 	       from_head - to_head);
 }
 
-static void unstuff_dinode(struct gfs2_inode *ip)
+void unstuff_dinode(struct gfs2_inode *ip)
 {
 	struct gfs2_sbd *sdp = ip->i_sbd;
 	struct gfs2_buffer_head *bh;
@@ -273,17 +287,22 @@ struct metapath *find_metapath(struct gfs2_inode *ip, uint64_t block)
 	uint64_t b = block;
 	unsigned int i;
 
-	zalloc(mp, sizeof(struct metapath));
-
-	for (i = ip->i_di.di_height; i--;)
-		mp->mp_list[i] = do_div(b, sdp->sd_inptrs);
+	mp = calloc(1, sizeof(struct metapath));
+	if (mp == NULL) {
+		fprintf(stderr, "Out of memory in %s\n", __FUNCTION__);
+		exit(-1);
+	}
+	for (i = ip->i_di.di_height; i--;) {
+		mp->mp_list[i] = b % sdp->sd_inptrs;
+		b /= sdp->sd_inptrs;
+	}
 
 	return mp;
 }
 
-static void lookup_block(struct gfs2_inode *ip,
-	     struct gfs2_buffer_head *bh, unsigned int height, struct metapath *mp,
-	     int create, int *new, uint64_t *block)
+void lookup_block(struct gfs2_inode *ip, struct gfs2_buffer_head *bh,
+		  unsigned int height, struct metapath *mp,
+		  int create, int *new, uint64_t *block)
 {
 	uint64_t *ptr = metapointer(bh, height, mp);
 
@@ -437,7 +456,8 @@ int gfs2_readi(struct gfs2_inode *ip, void *buf,
 
 	if (isdir) {
 		lblock = offset;
-		o = do_div(lblock, sdp->sd_jbsize);
+		o = lblock % sdp->sd_jbsize;
+		lblock /= sdp->sd_jbsize;
 	} else {
 		lblock = offset >> sdp->sd_sb.sb_bsize_shift;
 		o = offset & (sdp->bsize - 1);
@@ -496,7 +516,7 @@ int gfs2_writei(struct gfs2_inode *ip, void *buf,
 	uint32_t extlen = 0;
 	unsigned int amount;
 	int new;
-	int isdir = !!(S_ISDIR(ip->i_di.di_flags));
+	int isdir = !!(S_ISDIR(ip->i_di.di_mode));
 	const uint64_t start = offset;
 	int copied = 0;
 
@@ -509,7 +529,8 @@ int gfs2_writei(struct gfs2_inode *ip, void *buf,
 
 	if (isdir) {
 		lblock = offset;
-		o = do_div(lblock, sdp->sd_jbsize);
+		o = lblock % sdp->sd_jbsize;
+		lblock /= sdp->sd_jbsize;
 	} else {
 		lblock = offset >> sdp->sd_sb.sb_bsize_shift;
 		o = offset & (sdp->bsize - 1);
@@ -614,9 +635,8 @@ int gfs2_dirent_next(struct gfs2_inode *dip, struct gfs2_buffer_head *bh,
 	return 0;
 }
 
-static int
-dirent_alloc(struct gfs2_inode *dip, struct gfs2_buffer_head *bh, int name_len,
-			 struct gfs2_dirent **dent_out)
+static int dirent_alloc(struct gfs2_inode *dip, struct gfs2_buffer_head *bh,
+			int name_len, struct gfs2_dirent **dent_out)
 {
 	struct gfs2_dirent *dent, *new;
 	unsigned int rec_len = GFS2_DIRENT_SIZE(name_len);
@@ -646,6 +666,7 @@ dirent_alloc(struct gfs2_inode *dip, struct gfs2_buffer_head *bh, int name_len,
 	do {
 		uint16_t cur_rec_len;
 		uint16_t cur_name_len;
+		uint16_t new_rec_len;
 
 		cur_rec_len = be16_to_cpu(dent->de_rec_len);
 		cur_name_len = be16_to_cpu(dent->de_name_len);
@@ -661,8 +682,10 @@ dirent_alloc(struct gfs2_inode *dip, struct gfs2_buffer_head *bh, int name_len,
 				new->de_rec_len = cpu_to_be16(cur_rec_len -
 											  GFS2_DIRENT_SIZE(cur_name_len));
 				new->de_name_len = cpu_to_be16(name_len);
-				dent->de_rec_len = cpu_to_be16(cur_rec_len -
-											   be16_to_cpu(new->de_rec_len));
+
+				new_rec_len = be16_to_cpu(new->de_rec_len);
+				dent->de_rec_len = cpu_to_be16(cur_rec_len - new_rec_len);
+
 				*dent_out = new;
 				return 0;
 			}
@@ -678,10 +701,12 @@ dirent_alloc(struct gfs2_inode *dip, struct gfs2_buffer_head *bh, int name_len,
 }
 
 void dirent2_del(struct gfs2_inode *dip, struct gfs2_buffer_head *bh,
-				struct gfs2_dirent *prev, struct gfs2_dirent *cur)
+		 struct gfs2_dirent *prev, struct gfs2_dirent *cur)
 {
 	uint16_t cur_rec_len, prev_rec_len;
 
+	if (dip->i_di.di_entries)
+		dip->i_di.di_entries--;
 	if (!prev) {
 		cur->de_inum.no_formal_ino = 0;
 		return;
@@ -694,14 +719,14 @@ void dirent2_del(struct gfs2_inode *dip, struct gfs2_buffer_head *bh,
 	prev->de_rec_len = cpu_to_be16(prev_rec_len);
 }
 
-void gfs2_get_leaf_nr(struct gfs2_inode *dip, uint32_t index,
+void gfs2_get_leaf_nr(struct gfs2_inode *dip, uint32_t lindex,
 					  uint64_t *leaf_out)
 {
 	uint64_t leaf_no;
 	int count;
 
 	count = gfs2_readi(dip, (char *)&leaf_no,
-		      index * sizeof(uint64_t),
+		      lindex * sizeof(uint64_t),
 		      sizeof(uint64_t));
 	if (count != sizeof(uint64_t))
 		die("gfs2_get_leaf_nr:  Bad internal read.\n");
@@ -721,8 +746,7 @@ void gfs2_put_leaf_nr(struct gfs2_inode *dip, uint32_t inx, uint64_t leaf_out)
 		die("gfs2_put_leaf_nr:  Bad internal write.\n");
 }
 
-static void
-dir_split_leaf(struct gfs2_inode *dip, uint32_t index, uint64_t leaf_no)
+static void dir_split_leaf(struct gfs2_inode *dip, uint32_t lindex, uint64_t leaf_no)
 {
 	struct gfs2_buffer_head *nbh, *obh;
 	struct gfs2_leaf *nleaf, *oleaf;
@@ -752,10 +776,13 @@ dir_split_leaf(struct gfs2_inode *dip, uint32_t index, uint64_t leaf_no)
 	len = 1 << (dip->i_di.di_depth - be16_to_cpu(oleaf->lf_depth));
 	half_len = len >> 1;
 
-	start = (index & ~(len - 1));
-
-	zalloc(lp, half_len * sizeof(uint64_t));
+	start = (lindex & ~(len - 1));
 
+	lp = calloc(1, half_len * sizeof(uint64_t));
+	if (lp == NULL) {
+		fprintf(stderr, "Out of memory in %s\n", __FUNCTION__);
+		exit(-1);
+	}
 	count = gfs2_readi(dip, (char *)lp, start * sizeof(uint64_t),
 		      half_len * sizeof(uint64_t));
 	if (count != half_len * sizeof(uint64_t))
@@ -824,8 +851,7 @@ dir_split_leaf(struct gfs2_inode *dip, uint32_t index, uint64_t leaf_no)
 	brelse(nbh, updated);
 }
 
-static void
-dir_double_exhash(struct gfs2_inode *dip)
+static void dir_double_exhash(struct gfs2_inode *dip)
 {
 	struct gfs2_sbd *sdp = dip->i_sbd;
 	uint64_t *buf;
@@ -834,7 +860,11 @@ dir_double_exhash(struct gfs2_inode *dip)
 	int x;
 	int count;
 
-	zalloc(buf, 3 * sdp->sd_hash_bsize);
+	buf = calloc(1, 3 * sdp->sd_hash_bsize);
+	if (buf == NULL) {
+		fprintf(stderr, "Out of memory in %s\n", __FUNCTION__);
+		exit(-1);
+	}
 
 	for (block = dip->i_di.di_size >> sdp->sd_hash_bsize_shift; block--;) {
 		count = gfs2_readi(dip, (char *)buf,
@@ -895,12 +925,12 @@ int gfs2_get_leaf(struct gfs2_inode *dip, uint64_t leaf_no,
  * Returns: 0 on success, error code otherwise
  */
 
-static int get_first_leaf(struct gfs2_inode *dip, uint32_t index,
-						  struct gfs2_buffer_head **bh_out)
+static int get_first_leaf(struct gfs2_inode *dip, uint32_t lindex,
+			  struct gfs2_buffer_head **bh_out)
 {
 	uint64_t leaf_no;
 
-	gfs2_get_leaf_nr(dip, index, &leaf_no);
+	gfs2_get_leaf_nr(dip, lindex, &leaf_no);
 	*bh_out = bread(&dip->i_sbd->buf_list, leaf_no);
 	return 0;
 }
@@ -915,7 +945,7 @@ static int get_first_leaf(struct gfs2_inode *dip, uint32_t index,
  */
 
 static int get_next_leaf(struct gfs2_inode *dip,struct gfs2_buffer_head *bh_in,
-						 struct gfs2_buffer_head **bh_out)
+			 struct gfs2_buffer_head **bh_out)
 {
 	struct gfs2_leaf *leaf;
 
@@ -927,14 +957,13 @@ static int get_next_leaf(struct gfs2_inode *dip,struct gfs2_buffer_head *bh_in,
 	return 0;
 }
 
-static void
-dir_e_add(struct gfs2_inode *dip, char *filename, int len,
-		  struct gfs2_inum *inum, unsigned int type)
+static void dir_e_add(struct gfs2_inode *dip, const char *filename, int len,
+		      struct gfs2_inum *inum, unsigned int type)
 {
 	struct gfs2_buffer_head *bh, *nbh;
 	struct gfs2_leaf *leaf, *nleaf;
 	struct gfs2_dirent *dent;
-	uint32_t index;
+	uint32_t lindex;
 	uint32_t hash;
 	uint64_t leaf_no, bn;
 
@@ -942,11 +971,11 @@ dir_e_add(struct gfs2_inode *dip, char *filename, int len,
 	hash = gfs2_disk_hash(filename, len);
 	/* Have to kludge because (hash >> 32) gives hash for some reason. */
 	if (dip->i_di.di_depth)
-		index = hash >> (32 - dip->i_di.di_depth);
+		lindex = hash >> (32 - dip->i_di.di_depth);
 	else
-		index = 0;
+		lindex = 0;
 
-	gfs2_get_leaf_nr(dip, index, &leaf_no);
+	gfs2_get_leaf_nr(dip, lindex, &leaf_no);
 
 	for (;;) {
 		bh = bread(&dip->i_sbd->buf_list, leaf_no);
@@ -956,7 +985,7 @@ dir_e_add(struct gfs2_inode *dip, char *filename, int len,
 
 			if (be16_to_cpu(leaf->lf_depth) < dip->i_di.di_depth) {
 				brelse(bh, not_updated);
-				dir_split_leaf(dip, index, leaf_no);
+				dir_split_leaf(dip, lindex, leaf_no);
 				goto restart;
 
 			} else if (dip->i_di.di_depth < GFS2_DIR_MAX_DEPTH) {
@@ -1010,8 +1039,7 @@ dir_e_add(struct gfs2_inode *dip, char *filename, int len,
 	}
 }
 
-static void
-dir_make_exhash(struct gfs2_inode *dip)
+static void dir_make_exhash(struct gfs2_inode *dip)
 {
 	struct gfs2_sbd *sdp = dip->i_sbd;
 	struct gfs2_dirent *dent;
@@ -1048,7 +1076,8 @@ dir_make_exhash(struct gfs2_inode *dip)
 			break;
 	} while (gfs2_dirent_next(dip, bh, &dent) == 0);
 
-	dent->de_rec_len = cpu_to_be16(be16_to_cpu(dent->de_rec_len) +
+	dent->de_rec_len = be16_to_cpu(dent->de_rec_len);
+	dent->de_rec_len = cpu_to_be16(dent->de_rec_len +
 		sizeof(struct gfs2_dinode) - sizeof(struct gfs2_leaf));
 
 	brelse(bh, updated);
@@ -1069,8 +1098,8 @@ dir_make_exhash(struct gfs2_inode *dip)
 	dip->i_di.di_depth = y;
 }
 
-static void dir_l_add(struct gfs2_inode *dip, char *filename, int len,
-					  struct gfs2_inum *inum, unsigned int type)
+static void dir_l_add(struct gfs2_inode *dip, const char *filename, int len,
+		      struct gfs2_inum *inum, unsigned int type)
 {
 	struct gfs2_dirent *dent;
 
@@ -1089,8 +1118,8 @@ static void dir_l_add(struct gfs2_inode *dip, char *filename, int len,
 	dip->i_di.di_entries++;
 }
 
-void dir_add(struct gfs2_inode *dip, char *filename, int len,
-			 struct gfs2_inum *inum, unsigned int type)
+void dir_add(struct gfs2_inode *dip, const char *filename, int len,
+	     struct gfs2_inum *inum, unsigned int type)
 {
 	if (dip->i_di.di_flags & GFS2_DIF_EXHASH)
 		dir_e_add(dip, filename, len, inum, type);
@@ -1098,10 +1127,9 @@ void dir_add(struct gfs2_inode *dip, char *filename, int len,
 		dir_l_add(dip, filename, len, inum, type);
 }
 
-struct gfs2_buffer_head *
-init_dinode(struct gfs2_sbd *sdp, struct gfs2_inum *inum,
-	    unsigned int mode, uint32_t flags,
-	    struct gfs2_inum *parent)
+struct gfs2_buffer_head *init_dinode(struct gfs2_sbd *sdp,
+				     struct gfs2_inum *inum, unsigned int mode,
+				     uint32_t flags, struct gfs2_inum *parent)
 {
 	struct gfs2_buffer_head *bh;
 	struct gfs2_dinode di;
@@ -1161,8 +1189,8 @@ init_dinode(struct gfs2_sbd *sdp, struct gfs2_inum *inum,
 	return bh;
 }
 
-struct gfs2_inode *createi(struct gfs2_inode *dip, char *filename,
-						   unsigned int mode, uint32_t flags)
+struct gfs2_inode *createi(struct gfs2_inode *dip, const char *filename,
+			   unsigned int mode, uint32_t flags)
 {
 	struct gfs2_sbd *sdp = dip->i_sbd;
 	uint64_t bn;
@@ -1198,7 +1226,8 @@ struct gfs2_inode *createi(struct gfs2_inode *dip, char *filename,
  *
  * Returns: 1 if the files are the same, otherwise 0.
  */
-int gfs2_filecmp(const char *file1, const char *file2, int len_of_file2)
+
+static int gfs2_filecmp(const char *file1, const char *file2, int len_of_file2)
 {
 	if (strlen(file1) != len_of_file2)
 		return 0;
@@ -1216,11 +1245,10 @@ int gfs2_filecmp(const char *file1, const char *file2, int len_of_file2)
  *
  * Returns:
  */
-static int leaf_search(struct gfs2_inode *dip,
-					   struct gfs2_buffer_head *bh, 
-					   const char *filename, int len,
-                       struct gfs2_dirent **dent_out,
-					   struct gfs2_dirent **dent_prev)
+static int leaf_search(struct gfs2_inode *dip, struct gfs2_buffer_head *bh, 
+		       const char *filename, int len,
+		       struct gfs2_dirent **dent_out,
+		       struct gfs2_dirent **dent_prev)
 {
 	uint32_t hash;
 	struct gfs2_dirent *dent, *prev = NULL;
@@ -1274,14 +1302,14 @@ static int leaf_search(struct gfs2_inode *dip,
  *
  * Returns: 0 on sucess, error code otherwise
  */
-static int linked_leaf_search(struct gfs2_inode *dip,
-							  const char *filename, int len,
-                              struct gfs2_dirent **dent_out,
-							  struct gfs2_dirent **dent_prev,
-							  struct gfs2_buffer_head **bh_out)
+
+static int linked_leaf_search(struct gfs2_inode *dip, const char *filename,
+			      int len, struct gfs2_dirent **dent_out,
+			      struct gfs2_dirent **dent_prev,
+			      struct gfs2_buffer_head **bh_out)
 {
 	struct gfs2_buffer_head *bh = NULL, *bh_next;
-	uint32_t hsize, index;
+	uint32_t hsize, lindex;
 	uint32_t hash;
 	int error = 0;
 
@@ -1292,9 +1320,9 @@ static int linked_leaf_search(struct gfs2_inode *dip,
 	/*  Figure out the address of the leaf node.  */
 
 	hash = gfs2_disk_hash(filename, len);
-	index = hash >> (32 - dip->i_di.di_depth);
+	lindex = hash >> (32 - dip->i_di.di_depth);
 
-	error = get_first_leaf(dip, index, &bh_next);
+	error = get_first_leaf(dip, lindex, &bh_next);
 	if (error)
 		return error;
 
@@ -1320,7 +1348,7 @@ static int linked_leaf_search(struct gfs2_inode *dip,
 		}
 		
 		error = get_next_leaf(dip, bh, &bh_next);
-	}while (!error);
+	} while (!error);
 	
 	brelse(bh, not_updated);
 	
@@ -1336,7 +1364,7 @@ static int linked_leaf_search(struct gfs2_inode *dip,
  * Returns:
  */
 static int dir_e_search(struct gfs2_inode *dip, const char *filename,
-						int len, unsigned int *type, struct gfs2_inum *inum)
+			int len, unsigned int *type, struct gfs2_inum *inum)
 {
 	struct gfs2_buffer_head *bh = NULL;
 	struct gfs2_dirent *dent;
@@ -1365,7 +1393,7 @@ static int dir_e_search(struct gfs2_inode *dip, const char *filename,
  * Returns:
  */
 static int dir_l_search(struct gfs2_inode *dip, const char *filename,
-						int len, unsigned int *type, struct gfs2_inum *inum)
+			int len, unsigned int *type, struct gfs2_inum *inum)
 {
 	struct gfs2_buffer_head *dibh;
 	struct gfs2_dirent *dent;
@@ -1398,8 +1426,8 @@ static int dir_l_search(struct gfs2_inode *dip, const char *filename,
  *
  * Returns: 0 if found, -1 on failure, -ENOENT if not found.
  */
-int dir_search(struct gfs2_inode *dip, const char *filename, int len,
-			   unsigned int *type, struct gfs2_inum *inum)
+static int dir_search(struct gfs2_inode *dip, const char *filename, int len,
+		      unsigned int *type, struct gfs2_inum *inum)
 {
 	int error;
 
@@ -1416,17 +1444,17 @@ int dir_search(struct gfs2_inode *dip, const char *filename, int len,
 
 static int dir_e_del(struct gfs2_inode *dip, const char *filename, int len)
 {
-	int index;
+	int lindex;
 	int error;
 	int found = 0;
 	uint64_t leaf_no;
 	struct gfs2_buffer_head *bh = NULL;
 	struct gfs2_dirent *cur, *prev;
 
-	index = (1 << (dip->i_di.di_depth))-1;
+	lindex = (1 << (dip->i_di.di_depth))-1;
 
-	for(; (index >= 0) && !found; index--){
-		gfs2_get_leaf_nr(dip, index, &leaf_no);
+	for(; (lindex >= 0) && !found; lindex--){
+		gfs2_get_leaf_nr(dip, lindex, &leaf_no);
 
 		while(leaf_no && !found){
 			bh = bread(&dip->i_sbd->buf_list, leaf_no);
@@ -1454,7 +1482,7 @@ static int dir_e_del(struct gfs2_inode *dip, const char *filename, int len)
 }
 
 static int dir_l_del(struct gfs2_inode *dip, struct gfs2_buffer_head *dibh,
-					 const char *filename, int len){
+		     const char *filename, int len){
 	int error=0;
 	int got_buf = 0;
 	struct gfs2_dirent *cur, *prev;
@@ -1498,7 +1526,7 @@ static int dir_l_del(struct gfs2_inode *dip, struct gfs2_buffer_head *dibh,
  * Returns: 0 on success (or if it doesn't already exist), -1 on failure
  */
 int gfs2_dirent_del(struct gfs2_inode *dip, struct gfs2_buffer_head *bh,
-					const char *filename, int len){
+		    const char *filename, int len){
 	int error;
 
 	if(!S_ISDIR(dip->i_di.di_mode))
@@ -1521,7 +1549,7 @@ int gfs2_dirent_del(struct gfs2_inode *dip, struct gfs2_buffer_head *bh,
  * Returns: 0 on success, -EXXXX on failure
  */
 int gfs2_lookupi(struct gfs2_inode *dip, const char *filename, int len,
-				 struct gfs2_inode **ipp)
+		 struct gfs2_inode **ipp)
 {
 	struct gfs2_sbd *sdp = dip->i_sbd;
 	int error = 0;
@@ -1567,32 +1595,59 @@ void gfs2_free_block(struct gfs2_sbd *sdp, uint64_t block)
  * gfs2_freedi - unlink a disk inode by block number.
  * Note: currently only works for regular files.
  */
-int gfs2_freedi(struct gfs2_sbd *sdp, uint64_t block)
+int gfs2_freedi(struct gfs2_sbd *sdp, uint64_t diblock)
 {
 	struct gfs2_inode *ip;
-	struct gfs2_buffer_head *bh;
-	int x;
-	uint64_t p;
-	unsigned char *buf;
+	struct gfs2_buffer_head *bh, *nbh;
+	int h, head_size;
+	uint64_t *ptr, block;
 	struct rgrp_list *rgd;
-	
-	bh = bread(&sdp->buf_list, block);
+	uint32_t height;
+	osi_list_t metalist[GFS2_MAX_META_HEIGHT];
+	osi_list_t *cur_list, *next_list, *tmp;
+
+	for (h = 0; h < GFS2_MAX_META_HEIGHT; h++)
+		osi_list_init(&metalist[h]);
+
+	bh = bread(&sdp->buf_list, diblock);
 	ip = inode_get(sdp, bh);
-	if (ip->i_di.di_height > 0) {
-		buf = (unsigned char *)bh->b_data;
-		/* Free up all indirect blocks */
-		for (x = sizeof(struct gfs2_dinode); x < sdp->bsize;
-			 x += sizeof(uint64_t)) {
-			p = be64_to_cpu(*(uint64_t *)(buf + x));
-			if (p)
-				gfs2_free_block(sdp, p);
+	height = ip->i_di.di_height;
+	osi_list_add(&bh->b_altlist, &metalist[0]);
+
+	for (h = 0; h < height; h++){
+		cur_list = &metalist[h];
+		next_list = &metalist[h + 1];
+		head_size = (h > 0 ? sizeof(struct gfs2_meta_header) :
+			     sizeof(struct gfs2_dinode));
+
+		for (tmp = cur_list->next; tmp != cur_list; tmp = tmp->next){
+			bh = osi_list_entry(tmp, struct gfs2_buffer_head,
+					    b_altlist);
+
+			for (ptr = (uint64_t *)(bh->b_data + head_size);
+			     (char *)ptr < (bh->b_data + sdp->bsize); ptr++) {
+				if (!*ptr)
+					continue;
+
+				block = be64_to_cpu(*ptr);
+				gfs2_free_block(sdp, block);
+				if (h == height - 1) /* if not metadata */
+					continue; /* don't queue it up */
+				/* Read the next metadata block in the chain.
+				   First see if it's on the nvbuf_list. */
+				nbh = bfind(&sdp->nvbuf_list, block);
+				if (!nbh)
+					nbh = bread(&sdp->buf_list, block);
+				osi_list_add(&nbh->b_altlist, next_list);
+				brelse(nbh, not_updated);
+			}
 		}
 	}
 	/* Set the bitmap type for inode to free space: */
 	gfs2_set_bitmap(sdp, ip->i_di.di_num.no_addr, GFS2_BLKST_FREE);
 	inode_put(ip, updated);
 	/* Now we have to adjust the rg freespace count and inode count: */
-	rgd = gfs2_blk2rgrpd(sdp, block);
+	rgd = gfs2_blk2rgrpd(sdp, diblock);
 	/* The rg itself is in memory as rgd->rg, but there's most likely a  */
 	/* buffer in memory for the rg on disk because we used it to fix the */
 	/* bitmaps, some of which are on the same block on disk.             */
