@@ -15,7 +15,7 @@ static int client_size = 0;
 static struct client *client = NULL;
 static struct pollfd *pollfd = NULL;
 static char last_action[16];
-static int ccs_handle;
+int ccs_handle;
 
 struct client {
 	int fd;
@@ -30,7 +30,7 @@ static int do_read(int fd, void *buf, size_t count)
 	int rv, off = 0;
 
 	while (off < count) {
-		rv = read(fd, buf + off, count - off);
+		rv = read(fd, (char *)buf + off, count - off);
 		if (rv == 0)
 			return -1;
 		if (rv == -1 && errno == EINTR)
@@ -47,7 +47,7 @@ static int do_write(int fd, void *buf, size_t count)
 	int rv, off = 0;
 
  retry:
-	rv = write(fd, buf + off, count);
+	rv = write(fd, (char *)buf + off, count);
 	if (rv == -1 && errno == EINTR)
 		goto retry;
 	if (rv < 0) {
@@ -63,7 +63,7 @@ static int do_write(int fd, void *buf, size_t count)
 	return 0;
 }
 
-void read_ccs_name(char *path, char *name)
+void read_ccs_name(const char *path, char *name)
 {
 	char *str;
 	int error;
@@ -77,7 +77,7 @@ void read_ccs_name(char *path, char *name)
 	free(str);
 }
 
-void read_ccs_yesno(char *path, int *yes, int *no)
+void read_ccs_yesno(const char *path, int *yes, int *no)
 {
 	char *str;
 	int error;
@@ -98,7 +98,7 @@ void read_ccs_yesno(char *path, int *yes, int *no)
 	free(str);
 }
 
-void read_ccs_int(char *path, int *config_val)
+void read_ccs_int(const char *path, int *config_val)
 {
 	char *str;
 	int val;
@@ -124,17 +124,15 @@ void read_ccs_int(char *path, int *config_val)
 #define GROUPD_WAIT_PATH "/cluster/group/@groupd_wait"
 #define GROUPD_MODE_DELAY_PATH "/cluster/group/@groupd_mode_delay"
 
-int setup_ccs(void)
+static int setup_ccs(void)
 {
-	int i = 0, cd;
+	int cd;
 
-	while ((cd = ccs_connect()) < 0) {
-		sleep(1);
-		if (++i > 9 && !(i % 10))
-			log_print("connect to ccs error %d, "
-				  "check cluster status", cd);
+	cd = ccs_connect();
+	if (cd < 0) {
+		log_print("ccs_connect error %d %d", cd, errno);
+		return -1;
 	}
-
 	ccs_handle = cd;
 
 	/* These config values are set from cluster.conf only if they haven't
@@ -152,7 +150,7 @@ int setup_ccs(void)
 	return 0;
 }
 
-void close_ccs(void)
+static void close_ccs(void)
 {
 	ccs_disconnect(ccs_handle);
 }
@@ -299,7 +297,7 @@ enum {
 	DO_GET_VERSION,
 };
 
-int get_action(char *buf)
+static int get_action(char *buf)
 {
 	char act[16];
 	int i;
@@ -443,7 +441,7 @@ static void copy_group_data(group_t *g, group_data_t *data)
 	data->id = g->global_id;
 
 	if (g->app && g->app->current_event) {
-		event_t *ev = g->app->current_event;
+		ev = g->app->current_event;
 		data->event_state = ev->state;
 		data->event_nodeid = ev->nodeid;
 		data->event_id = ev->id;
@@ -699,7 +697,7 @@ static void process_listener(int ci)
 	log_debug("client connection %d fd %d", i, fd);
 }
 
-static int setup_listener(char *sock_path)
+static int setup_listener(const char *sock_path)
 {
 	struct sockaddr_un addr;
 	socklen_t addrlen;
@@ -736,8 +734,10 @@ static int setup_listener(char *sock_path)
 
 void cluster_dead(int ci)
 {
-	log_print("cluster is down, exiting");
+	if (!cluster_down)
+		log_print("cluster is down, exiting");
 	daemon_quit = 1;
+	cluster_down = 1;
 }
 
 #define min(x, y) ({                            \
@@ -769,12 +769,16 @@ static void loop(void)
 
 	setup_logging();
 
-	if (cfgd_groupd_compat == 0)
+	if (cfgd_groupd_compat == 0) {
 		group_mode = GROUP_LIBCPG;
-	else if (cfgd_groupd_compat == 1)
+		log_level(LOG_INFO, "groupd compatibility mode 0 cfg");
+	} else if (cfgd_groupd_compat == 1) {
 		group_mode = GROUP_LIBGROUP;
-	else if (cfgd_groupd_compat == 2)
+		log_level(LOG_INFO, "groupd compatibility mode 1 cfg");
+	} else if (cfgd_groupd_compat == 2) {
 		group_mode = GROUP_PENDING;
+		/* report the mode in set_group_mode() */
+	}
 
 	rv = setup_cpg();
 	if (rv < 0)
@@ -826,6 +830,7 @@ static void loop(void)
 		}
 	}
  out:
+	close_cpg();
 	close_ccs();
 	close_cman();
 
@@ -885,8 +890,8 @@ static void print_usage(void)
 	printf("\n");
 	printf("Options:\n");
 	printf("\n");
-	printf("  -D	       Enable debugging code and don't fork\n");
-	printf("  -L <num>     Enable (1) or disable (0) debugging to logsys (default %d)\n", DEFAULT_DEBUG_LOGSYS);
+	printf("  -D	       Enable debugging to stderr and don't fork\n");
+	printf("  -L           Enable debugging to log file\n");
 	printf("  -g <num>     group compatibility mode, 0 off, 1 on, 2 detect\n");
 	printf("               0: use libcpg, no backward compat, best performance\n");
 	printf("               1: use libgroup for compat with cluster2/rhel5\n");
@@ -903,7 +908,7 @@ static void print_usage(void)
 	printf("  -V	       Print program version information, then exit\n");
 }
 
-#define OPTION_STRING "L:Dg:w:d:hVv"
+#define OPTION_STRING "LDg:w:d:hVv"
 
 static void read_arguments(int argc, char **argv)
 {
@@ -920,8 +925,8 @@ static void read_arguments(int argc, char **argv)
 			break;
 
 		case 'L':
-			optd_debug_logsys = 1;
-			cfgd_debug_logsys = atoi(optarg);
+			optd_debug_logfile = 1;
+			cfgd_debug_logfile = 1;
 			break;
 
 		case 'g':
@@ -972,9 +977,9 @@ static void read_arguments(int argc, char **argv)
 		};
 	}
 
-	if (!optd_debug_logsys && getenv("GROUPD_DEBUG")) {
-		optd_debug_logsys = 1;
-		cfgd_debug_logsys = atoi(getenv("GROUPD_DEBUG"));
+	if (getenv("GROUPD_DEBUG")) {
+		optd_debug_logfile = 1;
+		cfgd_debug_logfile = 1;
 	}
 }
 
@@ -1018,7 +1023,6 @@ int main(int argc, char *argv[])
 		INIT_LIST_HEAD(&gd_levels[i]);
 
 	read_arguments(argc, argv);
-	lockfile();
 
 	if (!daemon_debug_opt) {
 		if (daemon(0, 0) < 0) {
@@ -1026,13 +1030,16 @@ int main(int argc, char *argv[])
 			exit(EXIT_FAILURE);
 		}
 	}
+	lockfile();
 	init_logging();
+	log_level(LOG_INFO, "groupd %s", RELEASE_VERSION);
 	signal(SIGTERM, sigterm_handler);
 	set_scheduler();
 	set_oom_adj(-16);
 
 	loop();
 
+	unlink(LOCKFILE_NAME);
 	return 0;
 }
 
@@ -1055,6 +1062,7 @@ void daemon_dump_save(void)
 int daemon_debug_opt;
 int daemon_debug_verbose;
 int daemon_quit;
+int cluster_down;
 int cman_quorate;
 int our_nodeid;
 char *our_name;
@@ -1070,10 +1078,10 @@ int group_mode;
 int optd_groupd_compat;
 int optd_groupd_wait;
 int optd_groupd_mode_delay;
-int optd_debug_logsys;
+int optd_debug_logfile;
 
 int cfgd_groupd_compat     = DEFAULT_GROUPD_COMPAT;
 int cfgd_groupd_wait       = DEFAULT_GROUPD_WAIT;
 int cfgd_groupd_mode_delay = DEFAULT_GROUPD_MODE_DELAY;
-int cfgd_debug_logsys      = DEFAULT_DEBUG_LOGSYS;
+int cfgd_debug_logfile     = DEFAULT_DEBUG_LOGFILE;
 
