@@ -23,7 +23,6 @@
 #define OP_LEAVE			5
 #define OP_JOINLEAVE			6
 
-static char *prog_name;
 static char *fsname;
 static int operation;
 static int opt_ind;
@@ -36,7 +35,7 @@ struct gfsc_mountgroup mgs[MAX_MG];
 struct gfsc_node nodes[MAX_NODES];
 
 
-static void print_usage(void)
+static void print_usage(const char *prog_name)
 {
 	printf("Usage:\n");
 	printf("\n");
@@ -64,13 +63,13 @@ static void decode_arguments(int argc, char **argv)
 			break;
 
 		case 'h':
-			print_usage();
+			print_usage(argv[0]);
 			exit(EXIT_SUCCESS);
 			break;
 
 		case 'V':
 			printf("%s %s (built %s %s)\n",
-				prog_name, RELEASE_VERSION, __DATE__, __TIME__);
+				argv[0], RELEASE_VERSION, __DATE__, __TIME__);
 			/* printf("%s\n", REDHAT_COPYRIGHT); */
 			exit(EXIT_SUCCESS);
 			break;
@@ -124,7 +123,7 @@ static void decode_arguments(int argc, char **argv)
 	}
 
 	if (!operation || !opt_ind) {
-		print_usage();
+		print_usage(argv[0]);
 		exit(EXIT_FAILURE);
 	}
 
@@ -141,7 +140,7 @@ static int do_write(int fd, void *buf, size_t count)
 	int rv, off = 0;
 
  retry:
-	rv = write(fd, buf + off, count);
+	rv = write(fd, (char *)buf + off, count);
 	if (rv == -1 && errno == EINTR)
 		goto retry;
 	if (rv < 0)
@@ -155,7 +154,7 @@ static int do_write(int fd, void *buf, size_t count)
 	return 0;
 }
 
-void do_leave(char *table)
+static void do_leave(char *table)
 {
 	struct gfsc_mount_args ma;
 	int rv;
@@ -169,7 +168,7 @@ void do_leave(char *table)
 		fprintf(stderr, "gfs_controld leave error %d\n", rv);
 }
 
-char *mg_flags_str(uint32_t flags)
+static char *mg_flags_str(uint32_t flags)
 {
 	static char str[128];
 	int i = 0;
@@ -214,7 +213,7 @@ char *mg_flags_str(uint32_t flags)
 	return str;
 }
 
-char *node_mount_str(uint32_t flags)
+static char *node_mount_str(uint32_t flags)
 {
 	static char str[128];
 	int i = 0;
@@ -246,7 +245,7 @@ char *node_mount_str(uint32_t flags)
 	return str;
 }
 
-int member_int(struct gfsc_node *n)
+static int member_int(struct gfsc_node *n)
 {
 	if (n->flags & GFSC_NF_DISALLOWED)
 		return -1;
@@ -255,7 +254,7 @@ int member_int(struct gfsc_node *n)
 	return 0;
 }
 
-char *condition_str(int cond)
+static const char *condition_str(int cond)
 {
 	switch (cond) {
 	case 0:
@@ -273,9 +272,9 @@ char *condition_str(int cond)
 	}
 }
 
-static void show_nodeids(int count, struct gfsc_node *nodes)
+static void show_nodeids(int count, struct gfsc_node *nodes_in)
 {
-	struct gfsc_node *n = nodes;
+	struct gfsc_node *n = nodes_in;
 	int i;
 
 	for (i = 0; i < count; i++) {
@@ -346,9 +345,9 @@ static void show_mg(struct gfsc_mountgroup *mg)
 	show_nodeids(node_count, nodes);
 }
 
-static void show_all_nodes(int count, struct gfsc_node *nodes)
+static void show_all_nodes(int count, struct gfsc_node *nodes_in)
 {
-	struct gfsc_node *n = nodes;
+	struct gfsc_node *n = nodes_in;
 	int i;
 
 	for (i = 0; i < count; i++) {
@@ -376,16 +375,15 @@ static void do_list(char *name)
 	memset(mgs, 0, sizeof(mgs));
 
 	if (name) {
-		rv = gfsc_mountgroup_info(name, mgs);
-		if (rv < 0)
-			return;
 		mg_count = 1;
+		rv = gfsc_mountgroup_info(name, mgs);
 	} else {
 		rv = gfsc_mountgroups(MAX_MG, &mg_count, mgs);
-		if (rv < 0)
-			return;
 	}
 
+	if (rv < 0)
+		exit(EXIT_FAILURE); /* gfs_controld probably not running */
+
 	if (mg_count)
 		printf("gfs mountgroups\n");
 
@@ -440,7 +438,6 @@ static void do_dump(void)
 
 int main(int argc, char **argv)
 {
-	prog_name = argv[0];
 	decode_arguments(argc, argv);
 
 	switch (operation) {
