@@ -1,5 +1,9 @@
 #include <assert.h>
 #include <platform.h>
+#include <time.h>
+#include <list.h>
+#include <restart_counter.h>
+#include <reslist.h>
 #include <message.h>
 #include <members.h>
 #include <stdio.h>
@@ -15,16 +19,9 @@
 #include <event.h>
 
 
-/*
- * Send a message to the target node to start the service.
- */
-int svc_start_remote(char *svcName, int request, uint32_t target);
-void svc_report_failure(char *);
-int get_service_state_internal(char *svcName, rg_state_t *svcStatus);
-
-
 /**
- *
+ * Send a message to the target node list, one at a time, to start
+ * the service.  Of course, the target list can be just 1 node.
  */
 int
 service_op_start(char *svcName,
@@ -72,19 +69,19 @@ service_op_start(char *svcName,
 			return RG_EFAIL;
 		default:
 			/* deliberate fallthrough */
-			log_printf(LOG_ERR,
+			logt_print(LOG_ERR,
 			       "#61: Invalid reply from member %d during"
 			       " start operation!\n", target);
 		case RG_NO:
 			/* state uncertain */
-			log_printf(LOG_CRIT, "State Uncertain: svc:%s "
+			logt_print(LOG_CRIT, "State Uncertain: svc:%s "
 			       "nid:%d req:%s ret:%d\n", svcName,
 			       target, rg_req_str(RG_START_REMOTE), ret);
 			return 0;
 		case 0:
 			if (new_owner)
 				*new_owner = target;
-			log_printf(LOG_NOTICE, "Service %s is now running "
+			logt_print(LOG_NOTICE, "Service %s is now running "
 			       "on member %d\n", svcName, (int)target);
 			return 0;
 		}
@@ -96,7 +93,7 @@ service_op_start(char *svcName,
 	else if (dep == target_list_len)
 		ret = RG_EDEPEND;
 
-	log_printf(LOG_INFO, "Start failed; node reports: %d failures, "
+	logt_print(LOG_INFO, "Start failed; node reports: %d failures, "
 	       "%d exclusive, %d dependency errors\n", fail, excl, dep);
 	return ret;
 }
@@ -146,7 +143,7 @@ service_op_stop(char *svcName, int do_disable, int event_type)
 	}
 
 	if (msg_open(MSG_CLUSTER, msgtarget, RG_PORT, &ctx, 2)< 0) {
-		log_printf(LOG_ERR,
+		logt_print(LOG_ERR,
 		       "#58: Failed opening connection to member #%d\n",
 		       my_id());
 		return -1;
@@ -158,7 +155,7 @@ service_op_stop(char *svcName, int do_disable, int event_type)
 	/* Send stop message to the other node */
 	if (msg_send(&ctx, &msg, sizeof (SmMessageSt)) < 
 	    (int)sizeof (SmMessageSt)) {
-		log_printf(LOG_ERR, "Failed to send complete message\n");
+		logt_print(LOG_ERR, "Failed to send complete message\n");
 		msg_close(&ctx);
 		return -1;
 	}
@@ -174,7 +171,7 @@ service_op_stop(char *svcName, int do_disable, int event_type)
 	} while(1);
 
 	if (msg_ret != sizeof (SmMessageSt)) {
-		log_printf(LOG_WARNING, "Strange response size: %d vs %d\n",
+		logt_print(LOG_WARNING, "Strange response size: %d vs %d\n",
 		       msg_ret, (int)sizeof(SmMessageSt));
 		return 0;	/* XXX really UNKNOWN */
 	}
@@ -190,7 +187,88 @@ service_op_stop(char *svcName, int do_disable, int event_type)
 
 
 /*
-   TODO
-   service_op_migrate()
+   service_op_migrate() - send a virtual machine to another host
+   in the cluster
  */
+int
+service_op_migrate(char *svcName,
+		   int target_node)
+{
+	SmMessageSt msg;
+	int msg_ret;
+	msgctx_t ctx;
+	rg_state_t svcStatus;
+	int msgtarget = my_id();
 
+	/* Build the message header */
+	msg.sm_hdr.gh_magic = GENERIC_HDR_MAGIC;
+	msg.sm_hdr.gh_command = RG_ACTION_REQUEST;
+	msg.sm_hdr.gh_arg1 = RG_ACTION_MASTER; 
+	msg.sm_hdr.gh_length = sizeof (SmMessageSt);
+
+	msg.sm_data.d_action = RG_MIGRATE;
+
+	strncpy(msg.sm_data.d_svcName, svcName,
+		sizeof(msg.sm_data.d_svcName));
+
+	msg.sm_data.d_ret = 0;
+	msg.sm_data.d_svcOwner = target_node;
+
+	/* Open a connection to the local node - it will decide what to
+	   do in this case. XXX inefficient; should queue requests
+	   locally and immediately forward requests otherwise */
+
+	if (get_service_state_internal(svcName, &svcStatus) < 0)
+		return RG_EFAIL;
+	if (svcStatus.rs_owner > 0) {
+		if (member_online(svcStatus.rs_owner)) {
+			msgtarget = svcStatus.rs_owner;
+		}
+
+		if (msgtarget <= 0) {
+			return RG_EFAIL;
+		}
+	}
+
+	if (msg_open(MSG_CLUSTER, msgtarget, RG_PORT, &ctx, 2)< 0) {
+		logt_print(LOG_ERR,
+		       "#58: Failed opening connection to member #%d\n",
+		       my_id());
+		return -1;
+	}
+
+	/* Encode */
+	swab_SmMessageSt(&msg);
+
+	/* Send stop message to the other node */
+	if (msg_send(&ctx, &msg, sizeof (SmMessageSt)) < 
+	    (int)sizeof (SmMessageSt)) {
+		logt_print(LOG_ERR, "Failed to send complete message\n");
+		msg_close(&ctx);
+		return -1;
+	}
+
+	/* Check the response */
+	do {
+		msg_ret = msg_receive(&ctx, &msg,
+				      sizeof (SmMessageSt), 10);
+		if ((msg_ret == -1 && errno != ETIMEDOUT) ||
+		    (msg_ret > 0)) {
+			break;
+		}
+	} while(1);
+
+	if (msg_ret != sizeof (SmMessageSt)) {
+		logt_print(LOG_WARNING, "Strange response size: %d vs %d\n",
+		       msg_ret, (int)sizeof(SmMessageSt));
+		return 0;	/* XXX really UNKNOWN */
+	}
+
+	/* Got a valid response from other node. */
+	msg_close(&ctx);
+
+	/* Decode */
+	swab_SmMessageSt(&msg);
+
+	return msg.sm_data.d_ret;
+}
